#include <iostream>
#include <chrono>

#include <opencv2/core.hpp>
#include <opencv2/videoio.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>

void test_passByValue(cv::Mat m);
void test_copy(const cv::Mat& m);
void test_clone(const cv::Mat& m);
void test_subMat(const cv::Mat& m);
void test_zoom(const cv::Mat& m);
void test_iterate(cv::Mat& m);
void test_colorspace(const cv::Mat& m);
void test_edges(const cv::Mat& m);
void test_affine(const cv::Mat& m);
void test_perspective(const cv::Mat& m);

int main(int argc, char* argv[])
{
	cv::VideoCapture capture;

	bool success = true;
	if (argc < 2)
	{
		success = capture.open(0);
	}
	else
	{
		success = capture.open(argv[1]);
		//capture.set(cv::CAP_PROP_FRAME_WIDTH, 640);
		//capture.set(cv::CAP_PROP_FRAME_HEIGHT, 480);
	}
	if (!success)
	{
		std::cerr << "Unable to open video capture" << std::endl;
		return 0;
	}

	cv::namedWindow("Input");
	cv::namedWindow("Processed");

	cv::Mat frame;
	while (capture.read(frame))
	{
		//test_passByValue(frame);
		//test_copy(frame);
		//test_clone(frame);
		//test_subMat(frame);
		//test_zoom(frame);
		test_iterate(frame);
		//test_colorspace(frame);
		//test_edges(frame);
		//test_affine(frame);
		//test_perspective(frame);

		cv::imshow("Input", frame);
		
		char key = static_cast<char>(cv::waitKey(1));
		if (key == 27) { std::cout << "Exiting..." << std::endl; break; }
	}

	return 0;
}

void test_passByValue(cv::Mat m)
{
	//m.setTo(cv::Scalar(255.0, 0.0, 0.0));
	//m.setTo(cv::Scalar(0.0, 255.0));
	//m.setTo(cv::Scalar(128.0));
	m.setTo(cv::Scalar::all(255.0));

	cv::imshow("Processed", m);
}

void test_copy(const cv::Mat& m)
{
	cv::Mat m_cpy = m;

	// invert colors, v1
	cv::Mat white(m.size(), m.type());
	//white.setTo(cv::Scalar(255.0, 255.0, 255.0));
	white = cv::Scalar::all(255.0);
	m_cpy = white - m_cpy;

	// invert colors, v2
	//m_clone = cv::Scalar::all(255.0) - m_clone;

	cv::imshow("Processed", m_cpy);
}

void test_clone(const cv::Mat& m)
{
	cv::Mat m_clone = m.clone();

	// invert colors, v3
	m_clone = ~m_clone;

	cv::Mat mask(m.size(), m.type());
	mask = cv::Scalar::all(0);
	cv::circle(mask, cv::Point(m.cols / 2, m.rows / 2), m.rows / 2, cv::Scalar::all(255), -1);

	m_clone = (mask & m_clone) + (~mask & m);

	cv::imshow("Processed", m_clone);
}

void test_subMat(const cv::Mat& m)
{
	cv::Rect zoomArea(m.cols / 2 - 200, m.rows / 2 - 100, 400, 200);
	cv::Mat subMat = m(zoomArea);
	//cv::Mat subMat = m(zoomArea).clone();

	cv::blur(subMat, subMat, cv::Size(20,20));

	cv::imshow("Processed", subMat);
}

void test_zoom(const cv::Mat& m)
{
	cv::Rect zoomArea(m.cols / 2 - 60, m.rows / 2 - 40, 120, 80);
	cv::Mat subMat = m(zoomArea);

	cv::Mat subMatZoomed;
	int zoomFactor = 2;
	cv::resize(subMat, subMatZoomed, subMat.size() * zoomFactor);
	subMatZoomed.copyTo(m(cv::Rect(0, 0, subMatZoomed.cols, subMatZoomed.rows)));

	cv::rectangle(subMat, cv::Rect(0, 0, subMat.cols, subMat.rows), cv::Scalar(0, 0, 255));

	cv::imshow("Processed", subMatZoomed);
}

// version 1: iterators
// http://docs.opencv.org/2.4/doc/tutorials/core/how_to_scan_images/how_to_scan_images.html#howtoscanimagesopencv
void iterate_v1(cv::Mat& m, unsigned char threshold)
{
	for (cv::MatIterator_<cv::Vec3b> it = m.begin<cv::Vec3b>(), end = m.end<cv::Vec3b>(); it != end; ++it)
	{
		(*it)[0] = (*it)[0] > threshold ? 255 : 0;
		(*it)[1] = (*it)[1] > threshold ? 255 : 0;
		(*it)[2] = (*it)[2] > threshold ? 255 : 0;
	}
}


// version 2: .at
void iterate_v2(cv::Mat& m, unsigned char threshold)
{
	for (int i = 0; i < m.rows; ++i)
	{
		for (int j = 0; j < m.cols; ++j)
		{
			cv::Vec3b pixel = m.at<cv::Vec3b>(i, j);
			pixel[0] = pixel[0] > threshold ? 255 : 0;
			pixel[1] = pixel[1] > threshold ? 255 : 0;
			pixel[2] = pixel[2] > threshold ? 255 : 0;
		}
	}
}

// version 3.1: pointers
// http://docs.opencv.org/2.4/doc/tutorials/core/how_to_scan_images/how_to_scan_images.html#howtoscanimagesopencv
void iterate_v3_1(cv::Mat& m, unsigned char threshold)
{
	int nChannels = m.channels();
	int nRows = m.rows, nCols = m.cols * nChannels;
	if (m.isContinuous())
	{
		nCols *= nRows;
		nRows = 1;
	}
	for (int i = 0; i < nRows; ++i)
	{
		uchar* p = m.ptr<uchar>(i);
		for (int j = 0; j < nCols; ++j)
		{
			p[j] = p[j] > threshold ? 255 : 0;
		}
	}
}

// version 3.2
void iterate_v3_2(cv::Mat& m, unsigned char threshold)
{
	CV_Assert(m.isContinuous());
	int nChannels = m.channels();
	int nRows = m.rows, nCols = m.cols * nChannels;
	uchar* p = m.data;
	for (int i = 0; i < nCols*nRows; ++i)
	{
		*p++ = *p > threshold ? 255 : 0;
	}
}

// version 4: foreach
void iterate_v4(cv::Mat& m, unsigned char threshold)
{
	m.forEach<cv::Vec3b>([&threshold](cv::Vec3b& p, const int * position) -> void {
		p[0] = p[0] > threshold ? 255 : 0;
		p[1] = p[1] > threshold ? 255 : 0;
		p[2] = p[2] > threshold ? 255 : 0;
	});
}

// version 5: OpenCV
void iterate_v5(cv::Mat& m, unsigned char threshold)
{
	cv::threshold(m, m, threshold, 255.0, cv::THRESH_BINARY);
}

void test_iterate(cv::Mat& m)
{
	CV_Assert(m.type() == CV_8UC3);

	unsigned char threshold = 128;
	std::vector<std::function<void(cv::Mat&, unsigned char)>> iterate_versions = {
		iterate_v1,
		iterate_v2,
		iterate_v3_1,
		iterate_v3_2,
		iterate_v4,
		iterate_v5
	};
	std::vector<std::string> functionString = { 
		"Version 1, iterator:\t", 
		"Version 2, .at:\t\t", 
		"Version 3.1, pointer:\t", 
		"Version 3.2, pointer:\t",
		"Version 4, foreach:\t",
		"Version 5, OpenCV:\t",
	};

	cv::Mat m_clone;
	for (size_t i = 0; i < iterate_versions.size(); ++i)
	{
		auto t_start = std::chrono::high_resolution_clock::now();
		m_clone = m.clone();

		iterate_versions[i](m_clone, threshold);

		auto t_end = std::chrono::high_resolution_clock::now();
		std::cout << functionString[i]
			<< std::chrono::duration_cast<std::chrono::microseconds>(t_end - t_start).count() << " usec, "
			<< std::chrono::duration_cast<std::chrono::milliseconds>(t_end - t_start).count() << " msec"
			<< std::endl;
	}
	std::cout << std::endl;

	cv::imshow("Processed", m_clone);
}

void test_colorspace(const cv::Mat & m)
{
	cv::Mat m_gray, m_hsv, m_xyz, m_rgb;

	cv::cvtColor(m, m_rgb, CV_BGR2RGB);
	cv::cvtColor(m, m_gray, CV_BGR2GRAY);
	cv::cvtColor(m, m_hsv, CV_BGR2HSV);
	cv::cvtColor(m, m_xyz, CV_BGR2XYZ);

	//m_hsv = m_hsv.mul(cv::Scalar(1, 1, 0.3));
	m_hsv = m_hsv.mul(cv::Scalar(1, 2, 1));

	cv::Mat result;
	cv::cvtColor(m_hsv, result, CV_HSV2BGR);

	cv::imshow("Processed", result);
}

void test_edges(const cv::Mat & m)
{
	cv::Mat gray, gray_blurred, grad_x, grad_y, grad, edgesAdded;
	cv::cvtColor(m, gray, CV_BGR2GRAY);

	cv::GaussianBlur(gray, gray_blurred, cv::Size(7, 7), 0, 0);

	int ddepth = CV_16S;
	cv::Sobel(gray_blurred, grad_x, ddepth, 1, 0, 3);
	cv::Sobel(gray_blurred, grad_y, ddepth, 0, 1, 3);
	cv::convertScaleAbs(grad_x, grad_x);
	cv::convertScaleAbs(grad_y, grad_y);
	grad = 0.5 * grad_x + 0.5* grad_y;

	//cv::Canny(gray, grad, 100, 200, 5);

	grad = 255 - grad;
	cv::threshold(grad, grad, 200, 255, cv::THRESH_BINARY);
	cv::cvtColor(grad, grad, CV_GRAY2BGR);
	edgesAdded = m.mul(grad, 1.0 / 255.0);

	cv::imshow("Processed", edgesAdded);
}

void test_affine(const cv::Mat & m)
{
	cv::Mat transformed;

	cv::Mat rotMat = cv::getRotationMatrix2D(cv::Point2f(m.cols / 2, m.rows / 2), 45.0, 0.5);
	cv::Mat translateMat = (cv::Mat_<float>(2,3) << 0.5, 0, 30, 0, 1, 100);
	cv::warpAffine(m, transformed, rotMat, m.size());
	//cv::warpAffine(m, transformed, translateMat, m.size());

	cv::imshow("Processed", transformed);
}

void test_perspective(const cv::Mat& m)
{
	cv::Mat transformed;

	std::vector<cv::Point2f> inputPoints = {
		cv::Point2f(0, 0),
		cv::Point2f(m.cols-1, 0),
		cv::Point2f(m.cols-1, m.rows-1),
		cv::Point2f(0, m.rows-1),
	};
	std::vector<cv::Point2f> outputPoints = {
		cv::Point2f(100, 50),
		cv::Point2f(m.cols - 100, 40),
		cv::Point2f(m.cols - 30, m.rows - 10),
		cv::Point2f(20, m.rows - 1),
	};
	cv::Mat transformMat = cv::getPerspectiveTransform(inputPoints, outputPoints);

	cv::warpPerspective(m, transformed, transformMat, m.size());

	cv::imshow("Processed", transformed);
}